# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import attempt_import
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.core.expr import ProductExpression, SumExpression
from pyomo.common.errors import InvalidValueError
import pyomo.common.unittest as unittest
from pyomo.contrib.solver.solvers.gurobi.gurobi_direct_minlp import (
    GurobiMINLPVisitor,
    GurobiDirectMINLP,
)
from pyomo.contrib.solver.tests.solvers.gurobi_to_pyomo_expressions import (
    grb_nl_to_pyo_expr,
)
from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    Integers,
    log,
    NonNegativeIntegers,
    NonNegativeReals,
    NonPositiveIntegers,
    NonPositiveReals,
    Objective,
    Param,
    Reals,
    sqrt,
    Var,
)

gurobipy, gurobipy_available = attempt_import('gurobipy', minimum_version='12.0.0')

if gurobipy_available:
    from gurobipy import GRB

    if not GurobiDirectMINLP().available():
        gurobipy_available = False


class CommonTest(unittest.TestCase):
    def get_model(self):
        m = ConcreteModel()
        m.x1 = Var(domain=NonNegativeReals)
        m.x2 = Var(domain=Reals)
        m.x3 = Var(domain=NonPositiveReals)
        m.y1 = Var(domain=Integers)
        m.y2 = Var(domain=NonNegativeIntegers)
        m.y3 = Var(domain=NonPositiveIntegers)
        m.z1 = Var(domain=Binary)

        return m

    def get_visitor(self):
        grb_model = gurobipy.Model()
        return GurobiMINLPVisitor(grb_model, symbolic_solver_labels=True)


@unittest.skipUnless(gurobipy_available, "gurobipy is not available")
@unittest.pytest.mark.solver("gurobi_direct_minlp")
class TestGurobiMINLPWalker(CommonTest):
    def _get_nl_expr_tree(self, visitor, expr):
        # This is a bit hacky, but the only way that I know to get the expression tree
        # publicly is from a general nonlinear constraint. So we can create it, and
        # then pull out the expression we just used to test it
        grb_model = visitor.grb_model
        aux = grb_model.addVar()
        grb_model.addConstr(aux == expr)
        grb_model.update()
        constrs = grb_model.getGenConstrs()
        self.assertEqual(len(constrs), 1)

        aux_var, opcode, data, parent = grb_model.getGenConstrNLAdv(constrs[0])
        self.assertIs(aux_var, aux)
        return opcode, data, parent

    def test_var_domains(self):
        m = self.get_model()
        e = m.x1 + m.x2 + m.x3 + m.y1 + m.y2 + m.y3 + m.z1
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(e)

        # We don't call update in walk expression for performance reasons, but
        # we need to update here in order to be able to test expr.
        visitor.grb_model.update()

        x1 = visitor.var_map[id(m.x1)]
        x2 = visitor.var_map[id(m.x2)]
        x3 = visitor.var_map[id(m.x3)]
        y1 = visitor.var_map[id(m.y1)]
        y2 = visitor.var_map[id(m.y2)]
        y3 = visitor.var_map[id(m.y3)]
        z1 = visitor.var_map[id(m.z1)]

        self.assertEqual(x1.lb, 0)
        self.assertEqual(x1.ub, float('inf'))
        self.assertEqual(x1.vtype, GRB.CONTINUOUS)

        self.assertEqual(x2.lb, -float('inf'))
        self.assertEqual(x2.ub, float('inf'))
        self.assertEqual(x2.vtype, GRB.CONTINUOUS)

        self.assertEqual(x3.lb, -float('inf'))
        self.assertEqual(x3.ub, 0)
        self.assertEqual(x3.vtype, GRB.CONTINUOUS)

        self.assertEqual(y1.lb, -float('inf'))
        self.assertEqual(y1.ub, float('inf'))
        self.assertEqual(y1.vtype, GRB.INTEGER)

        self.assertEqual(y2.lb, 0)
        self.assertEqual(y2.ub, float('inf'))
        self.assertEqual(y2.vtype, GRB.INTEGER)

        self.assertEqual(y3.lb, -float('inf'))
        self.assertEqual(y3.ub, 0)
        self.assertEqual(y3.vtype, GRB.INTEGER)

        self.assertEqual(z1.vtype, GRB.BINARY)

    def test_var_bounds(self):
        m = self.get_model()
        m.x2.setlb(-34)
        m.x2.setub(45)
        m.x3.setub(5)
        m.y1.setlb(-2)
        m.y1.setub(3)
        m.y2.setlb(-5)
        m.z1.setub(4)
        m.z1.setlb(-3)

        e = m.x1 + m.x2 + m.x3 + m.y1 + m.y2 + m.y3 + m.z1
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(e)

        # We don't call update in walk expression for performance reasons, but
        # we need to update here in order to be able to test expr.
        visitor.grb_model.update()

        x2 = visitor.var_map[id(m.x2)]
        x3 = visitor.var_map[id(m.x3)]
        y1 = visitor.var_map[id(m.y1)]
        y2 = visitor.var_map[id(m.y2)]
        z1 = visitor.var_map[id(m.z1)]

        self.assertEqual(x2.lb, -34)
        self.assertEqual(x2.ub, 45)
        self.assertEqual(x2.vtype, GRB.CONTINUOUS)

        self.assertEqual(x3.lb, -float('inf'))
        self.assertEqual(x3.ub, 0)
        self.assertEqual(x3.vtype, GRB.CONTINUOUS)

        self.assertEqual(y1.lb, -2)
        self.assertEqual(y1.ub, 3)
        self.assertEqual(y1.vtype, GRB.INTEGER)

        self.assertEqual(y2.lb, 0)
        self.assertEqual(y2.ub, float('inf'))
        self.assertEqual(y2.vtype, GRB.INTEGER)

        self.assertEqual(z1.vtype, GRB.BINARY)

    def test_write_addition(self):
        m = self.get_model()
        m.c = Constraint(expr=m.x1 + m.x2 >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        x1 = visitor.var_map[id(m.x1)]
        x2 = visitor.var_map[id(m.x2)]

        # This is a linear expression
        self.assertEqual(expr.size(), 2)
        self.assertEqual(expr.getCoeff(0), 1.0)
        self.assertEqual(expr.getCoeff(1), 1.0)
        self.assertIs(expr.getVar(0), x1)
        self.assertIs(expr.getVar(1), x2)
        self.assertEqual(expr.getConstant(), 0.0)

    def test_write_subtraction(self):
        m = self.get_model()
        m.c = Constraint(expr=m.x1 - m.x2 >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        x1 = visitor.var_map[id(m.x1)]
        x2 = visitor.var_map[id(m.x2)]

        # Also linear, whoot!
        self.assertEqual(expr.size(), 2)
        self.assertEqual(expr.getCoeff(0), 1.0)
        self.assertEqual(expr.getCoeff(1), -1.0)
        self.assertIs(expr.getVar(0), x1)
        self.assertIs(expr.getVar(1), x2)
        self.assertEqual(expr.getConstant(), 0.0)

    def test_write_product(self):
        m = self.get_model()
        m.c = Constraint(expr=m.x1 * m.x2 >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        x1 = visitor.var_map[id(m.x1)]
        x2 = visitor.var_map[id(m.x2)]

        # This is quadratic
        self.assertEqual(expr.size(), 1)
        lin_expr = expr.getLinExpr()
        self.assertEqual(lin_expr.size(), 0)
        self.assertIs(expr.getVar1(0), x1)
        self.assertIs(expr.getVar2(0), x2)
        self.assertEqual(expr.getCoeff(0), 1.0)

    def test_write_product_with_fixed_var(self):
        m = self.get_model()
        m.x2.fix(4)
        m.c = Constraint(expr=m.x1 * m.x2 == 1)

        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        x1 = visitor.var_map[id(m.x1)]

        # this is linear
        self.assertEqual(expr.size(), 1)
        self.assertEqual(expr.getCoeff(0), 4.0)
        self.assertIs(expr.getVar(0), x1)
        self.assertEqual(expr.getConstant(), 0.0)

    def test_write_product_with_0(self):
        m = self.get_model()
        m.c = Constraint(expr=(0 * m.x1 * m.x2) * m.x3 == 0)

        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # this is a "nonlinear"
        opcode, data, parent = self._get_nl_expr_tree(visitor, expr)

        reverse_var_map = {grb_v: pyo_v for pyo_v, grb_v in visitor.var_map.items()}
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)

        assertExpressionsEqual(
            self,
            pyo_expr,
            ProductExpression((ProductExpression((0.0, m.x1, m.x2, m.x3)),)),
        )

    def test_write_division(self):
        m = self.get_model()
        m.c = Constraint(expr=1 / m.x1 == 1)

        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        visitor.grb_model.update()
        grb_to_pyo_var_map = {
            grb_var: py_var for py_var, grb_var in visitor.var_map.items()
        }

        opcode, data, parent = self._get_nl_expr_tree(visitor, expr)

        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, grb_to_pyo_var_map)
        assertExpressionsEqual(self, pyo_expr, 1.0 / m.x1)

    def test_write_division_linear(self):
        m = self.get_model()
        m.p = Param(initialize=3, mutable=True)
        m.c = Constraint(expr=(m.x1 + m.x2) * m.p / 10 == 1)

        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        x1 = visitor.var_map[id(m.x1)]
        x2 = visitor.var_map[id(m.x2)]

        # linear
        self.assertEqual(expr.size(), 2)
        self.assertEqual(expr.getConstant(), 0)
        self.assertAlmostEqual(expr.getCoeff(0), 3 / 10)
        self.assertIs(expr.getVar(0), x1)
        self.assertAlmostEqual(expr.getCoeff(1), 3 / 10)
        self.assertIs(expr.getVar(1), x2)

    def test_write_linear_power_expression_var_const(self):
        m = self.get_model()
        m.devious = Param(initialize=1, mutable=True)
        m.c = Constraint(expr=m.x1**m.devious >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        x1 = visitor.var_map[id(m.x1)]

        # It's just a single var
        self.assertIs(expr, x1)
        self.assertEqual(len(visitor.grb_model.getGenConstrs()), 0)

        # now try a linear expression
        m.c2 = Constraint(expr=(m.x1 + 2 * m.x2) ** m.devious >= 5)
        _, lin_expr = visitor.walk_expression(m.c2.body)
        self.assertEqual(len(visitor.grb_model.getGenConstrs()), 0)
        x2 = visitor.var_map[m.x2]
        self.assertEqual(lin_expr.size(), 2)
        self.assertEqual(lin_expr.getConstant(), 0)
        self.assertIs(lin_expr.getVar(0), x1)
        self.assertIs(lin_expr.getVar(1), x2)
        self.assertEqual(lin_expr.getCoeff(0), 1.0)
        self.assertEqual(lin_expr.getCoeff(1), 2.0)

        # now do a quadratic expression
        m.c3 = Constraint(expr=(m.x1**2 + 5.4) ** m.devious >= 8)
        _, quad_expr = visitor.walk_expression(m.c3.body)
        self.assertEqual(len(visitor.grb_model.getGenConstrs()), 0)
        self.assertEqual(quad_expr.size(), 1)
        expr = quad_expr.getLinExpr()
        # no vars in linear part, just the constant
        self.assertEqual(expr.size(), 0)
        self.assertEqual(expr.getConstant(), 5.4)
        self.assertIs(quad_expr.getVar1(0), x1)
        self.assertIs(quad_expr.getVar2(0), x1)
        self.assertEqual(quad_expr.getCoeff(0), 1.0)

    def test_write_quadratic_power_expression_var_const(self):
        m = self.get_model()
        m.c = Constraint(expr=m.x1**2 >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # This is quadratic
        x1 = visitor.var_map[id(m.x1)]

        self.assertEqual(expr.size(), 1)
        lin_expr = expr.getLinExpr()
        self.assertEqual(lin_expr.size(), 0)
        self.assertEqual(lin_expr.getConstant(), 0)
        self.assertIs(expr.getVar1(0), x1)
        self.assertIs(expr.getVar2(0), x1)
        self.assertEqual(expr.getCoeff(0), 1.0)

    def test_write_quadratic_constant_pow_expression(self):
        m = self.get_model()
        m.c = Constraint(expr=(m.x1**2 + 2 * m.x2 + 3) ** 2 <= 7)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # This is general nonlinear
        opcode, data, parent = self._get_nl_expr_tree(visitor, expr)

        reverse_var_map = {grb_v: pyo_v for pyo_v, grb_v in visitor.var_map.items()}
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)

        assertExpressionsEqual(
            self,
            pyo_expr,
            SumExpression((3.0, ProductExpression((2.0, m.x2)), m.x1**2)) ** 2,
        )

    def test_write_nonquadratic_power_expression_var_const(self):
        m = self.get_model()
        m.c = Constraint(expr=m.x1**3 >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # This is general nonlinear
        opcode, data, parent = self._get_nl_expr_tree(visitor, expr)

        reverse_var_map = {grb_v: pyo_v for pyo_v, grb_v in visitor.var_map.items()}
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)

        assertExpressionsEqual(self, pyo_expr, m.x1**3.0)

    def test_write_power_expression_var_var(self):
        m = self.get_model()
        m.c = Constraint(expr=m.x1**m.x2 >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # You can't actually use this in a model in Gurobi 12, but you can build the
        # expression... (It fails during the solve.)

        x1 = visitor.var_map[id(m.x1)]
        x2 = visitor.var_map[id(m.x2)]

        opcode, data, parent = self._get_nl_expr_tree(visitor, expr)

        reverse_var_map = {grb_v: pyo_v for pyo_v, grb_v in visitor.var_map.items()}
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)

        assertExpressionsEqual(self, pyo_expr, m.x1**m.x2)

    def test_write_power_expression_const_var(self):
        m = self.get_model()
        m.c = Constraint(expr=2**m.x2 >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        x2 = visitor.var_map[id(m.x2)]

        opcode, data, parent = self._get_nl_expr_tree(visitor, expr)

        reverse_var_map = {grb_v: pyo_v for pyo_v, grb_v in visitor.var_map.items()}
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)

        assertExpressionsEqual(self, pyo_expr, 2.0**m.x2)

    def test_write_absolute_value_of_constant(self):
        m = self.get_model()
        m.tricky = Param(initialize=-3.4, mutable=True)
        m.c = Constraint(expr=abs(m.tricky + m.x2) + m.x1 <= 7)
        m.x2.fix(1)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        x1 = visitor.var_map[m.x1]

        self.assertEqual(len(visitor.grb_model.getGenConstrs()), 0)
        self.assertEqual(expr.size(), 1)
        self.assertEqual(expr.getConstant(), 2.4)
        self.assertIs(expr.getVar(0), x1)
        self.assertEqual(expr.getCoeff(0), 1.0)

    def test_write_absolute_value_of_var(self):
        # Gurobi doesn't support abs of expressions, so we have to do a factorable
        # programming thing...
        m = self.get_model()
        m.c = Constraint(expr=abs(m.x1) >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # expr is actually an auxiliary variable. We should
        # get a constraint:
        # expr == abs(x1)
        x1 = visitor.var_map[id(m.x1)]

        self.assertIsInstance(expr, gurobipy.Var)
        grb_model = visitor.grb_model
        # We don't call update in walk expression for performance reasons, but
        # we need to update here in order to be able to test expr.
        grb_model.update()
        self.assertEqual(grb_model.numVars, 2)
        self.assertEqual(grb_model.numGenConstrs, 1)
        self.assertEqual(grb_model.numConstrs, 0)
        self.assertEqual(grb_model.numQConstrs, 0)

        cons = grb_model.getGenConstrs()[0]
        aux, v = grb_model.getGenConstrAbs(cons)
        self.assertIs(aux, expr)
        self.assertIs(v, x1)

    def test_write_absolute_value_of_expression(self):
        m = self.get_model()
        m.c = Constraint(expr=abs(m.x1 + 2 * m.x2) >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # expr is actually an auxiliary variable. We should
        # get three constraints:
        # aux1 == x1 + 2 * x2
        # expr == abs(aux1)

        x1 = visitor.var_map[m.x1]
        x2 = visitor.var_map[m.x2]

        # we're going to have to write the resulting model to an lp file to test that we
        # have what we expect
        self.assertIsInstance(expr, gurobipy.Var)
        grb_model = visitor.grb_model
        # We don't call update in walk expression for performance reasons, but
        # we need to update here in order to be able to test expr.
        grb_model.update()
        self.assertEqual(grb_model.numVars, 4)
        self.assertEqual(grb_model.numGenConstrs, 1)
        self.assertEqual(grb_model.numConstrs, 1)
        self.assertEqual(grb_model.numQConstrs, 0)

        cons = grb_model.getGenConstrs()[0]
        aux2, aux1 = grb_model.getGenConstrAbs(cons)
        self.assertIs(aux2, expr)

        cons = grb_model.getConstrs()[0]
        # this guy is linear equality
        self.assertEqual(cons.RHS, 0)
        self.assertEqual(cons.Sense, '=')
        linexpr = grb_model.getRow(cons)
        self.assertEqual(linexpr.getConstant(), 0)
        self.assertEqual(linexpr.size(), 3)
        self.assertEqual(linexpr.getCoeff(0), -1)
        self.assertIs(linexpr.getVar(0), x1)
        self.assertEqual(linexpr.getCoeff(1), -2)
        self.assertIs(linexpr.getVar(1), x2)
        self.assertEqual(linexpr.getCoeff(2), 1)
        self.assertIs(linexpr.getVar(2), aux1)

    def test_write_expression_with_mutable_param(self):
        m = self.get_model()
        m.p = Param(initialize=4, mutable=True)
        m.c = Constraint(expr=m.p**m.x2 >= 3)
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # expr is nonlinear
        x2 = visitor.var_map[id(m.x2)]

        opcode, data, parent = self._get_nl_expr_tree(visitor, expr)

        reverse_var_map = {grb_v: pyo_v for pyo_v, grb_v in visitor.var_map.items()}
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)

        assertExpressionsEqual(self, pyo_expr, 4.0**m.x2)

    def test_monomial_expression(self):
        m = self.get_model()
        m.p = Param(initialize=4, mutable=True)

        const_expr = 3 * m.x1
        nested_expr = (1 / m.p) * m.x1
        pow_expr = (m.p ** (0.5)) * m.x1

        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(const_expr)
        x1 = visitor.var_map[id(m.x1)]
        self.assertEqual(expr.size(), 1)
        self.assertEqual(expr.getConstant(), 0.0)
        self.assertIs(expr.getVar(0), x1)
        self.assertEqual(expr.getCoeff(0), 3)

        _, expr = visitor.walk_expression(nested_expr)
        self.assertEqual(expr.size(), 1)
        self.assertEqual(expr.getConstant(), 0.0)
        self.assertIs(expr.getVar(0), x1)
        self.assertAlmostEqual(expr.getCoeff(0), 1 / 4)

        _, expr = visitor.walk_expression(pow_expr)
        self.assertEqual(expr.size(), 1)
        self.assertEqual(expr.getConstant(), 0.0)
        self.assertIs(expr.getVar(0), x1)
        self.assertEqual(expr.getCoeff(0), 2)

    def test_log_expression(self):
        m = self.get_model()
        m.c = Constraint(expr=log(m.x1) >= 3)
        m.pprint()
        visitor = self.get_visitor()
        _, expr = visitor.walk_expression(m.c.body)

        # expr is nonlinear
        x1 = visitor.var_map[id(m.x1)]

        opcode, data, parent = self._get_nl_expr_tree(visitor, expr)

        reverse_var_map = {grb_v: pyo_v for pyo_v, grb_v in visitor.var_map.items()}
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)

        assertExpressionsEqual(self, pyo_expr, log(m.x1))

    def test_handle_complex_number_sqrt(self):
        m = self.get_model()
        m.p = Param(initialize=3, mutable=True)
        m.c = Constraint(expr=sqrt(-m.p) + m.x1 >= 3)

        visitor = self.get_visitor()
        with self.assertRaisesRegex(
            InvalidValueError,
            r"Invalid number encountered evaluating constant unary expression "
            r"sqrt\(\s*-\s*p\s*\): "
            r"(?:math domain error|expected a nonnegative input, got -?\d+(?:\.\d+)?)",
        ):
            _, expr = visitor.walk_expression(m.c.body)

    def test_handle_invalid_log(self):
        m = self.get_model()
        m.p = Param(initialize=0, mutable=True)
        m.c = Constraint(expr=log(m.p) + m.x1 >= 3)

        visitor = self.get_visitor()
        with self.assertRaisesRegex(
            InvalidValueError,
            r"Invalid number encountered evaluating constant unary expression "
            r"log\(\s*p\s*\): "
            r"(?:math domain error|expected a positive input)",
        ):
            _, expr = visitor.walk_expression(m.c.body)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import ConcreteModel, Var, Objective, Integers
from pyomo.opt import TerminationCondition
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class MILP_unbounded(_BaseTestModel):
    """
    An unbounded mixed-integer linear program
    """

    description = "MILP_unbounded"
    capabilities = set(['linear', 'integer'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.solve_should_fail = True
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.x = Var(within=Integers)
        model.y = Var(within=Integers)

        model.o = Objective(expr=model.x + model.y)

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.x.value = None
        model.y.value = None

    def post_solve_test_validation(self, tester, results):
        if tester is None:
            assert results['Solver'][0]['termination condition'] in (
                TerminationCondition.unbounded,
                TerminationCondition.infeasibleOrUnbounded,
            )
        else:
            tester.assertIn(
                results['Solver'][0]['termination condition'],
                (
                    TerminationCondition.unbounded,
                    TerminationCondition.infeasibleOrUnbounded,
                ),
            )


@register_model
class MILP_unbounded_kernel(MILP_unbounded):
    def _generate_model(self):
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.x = pmo.variable(domain=pmo.IntegerSet)
        model.y = pmo.variable(domain=pmo.IntegerSet)

        model.o = pmo.objective(model.x + model.y)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Example in paper 'Using regularization and second order information in outer approximation for convex MINLP'

Ref:
Kronqvist J, Bernal D E, Grossmann I E. Using regularization and second order information in outer approximation for convex MINLP[J]. Mathematical Programming, 2020, 180(1): 285-310.

Problem type:   nonconvex MINLP
        size:   1  binary variable
                1  continuous variables
                3  constraints

"""

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    Reals,
    Integers,
    Objective,
    Var,
    minimize,
    exp,
)
from pyomo.common.collections import ComponentMap


class SimpleMINLP5(ConcreteModel):
    def __init__(self, *args, **kwargs):
        """Create the problem."""
        kwargs.setdefault('name', 'SimpleMINLP5')
        super(SimpleMINLP5, self).__init__(*args, **kwargs)
        m = self

        m.x = Var(within=Reals, bounds=(1, 20), initialize=5.29)
        m.y = Var(within=Integers, bounds=(1, 20), initialize=3)

        m.objective = Objective(
            expr=0.3 * (m.x - 8) ** 2
            + 0.04 * (m.y - 6) ** 4
            + 0.1 * exp(2 * m.x) * (m.y ** (-4)),
            sense=minimize,
        )

        m.c1 = Constraint(expr=6 * m.x + m.y <= 60)
        m.c2 = Constraint(expr=1 / m.x + 1 / m.x - m.x**0.5 * m.y**0.5 <= -1)
        m.c3 = Constraint(expr=2 * m.x - 5 * m.y <= -1)
        m.optimal_value = 3.6572
        m.optimal_solution = ComponentMap()
        m.optimal_solution[m.x] = 4.991797840270567
        m.optimal_solution[m.y] = 7.0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io
from contextlib import redirect_stdout

import pyomo.common.unittest as unittest
import pyomo.environ as pyo

from pyomo.contrib.solver.common.results import SolutionStatus, TerminationCondition
from pyomo.contrib.solver.solvers.knitro.api import knitro
from pyomo.contrib.solver.solvers.knitro.config import KnitroConfig
from pyomo.contrib.solver.solvers.knitro.direct import KnitroDirectSolver
from pyomo.contrib.solver.solvers.knitro.engine import Engine

avail = KnitroDirectSolver().available()


@unittest.skipIf(not avail, "KNITRO solver is not available")
@unittest.pytest.mark.solver("knitro_direct")
class TestKnitroDirectSolverConfig(unittest.TestCase):
    def test_default_instantiation(self):
        config = KnitroConfig()
        self.assertIsNone(config._description)
        self.assertEqual(config._visibility, 0)
        self.assertFalse(config.tee)
        self.assertTrue(config.load_solutions)
        self.assertTrue(config.raise_exception_on_nonoptimal_result)
        self.assertFalse(config.symbolic_solver_labels)
        self.assertIsNone(config.timer)
        self.assertIsNone(config.threads)
        self.assertIsNone(config.time_limit)

    def test_custom_instantiation(self):
        config = KnitroConfig(description="A description")
        config.tee = True
        self.assertTrue(config.tee)
        self.assertEqual(config._description, "A description")
        self.assertIsNone(config.time_limit)


@unittest.skipIf(not avail, "KNITRO solver is not available")
@unittest.pytest.mark.solver("knitro_direct")
class TestKnitroSolverResultsExtraInfo(unittest.TestCase):
    def test_results_extra_info_mip(self):
        """Test that MIP-specific extra info is populated for MIP problems."""
        opt = KnitroDirectSolver()
        m = pyo.ConcreteModel()
        m.x = pyo.Var(domain=pyo.Integers, bounds=(0, 10))
        m.y = pyo.Var(domain=pyo.Integers, bounds=(0, 10))
        m.obj = pyo.Objective(expr=m.x + m.y, sense=pyo.maximize)
        m.c1 = pyo.Constraint(expr=2 * m.x + m.y <= 15)
        m.c2 = pyo.Constraint(expr=m.x + 2 * m.y <= 15)

        results = opt.solve(m)

        # Check that MIP-specific fields are populated
        self.assertIsNotNone(results.extra_info.mip_number_nodes)
        self.assertIsNotNone(results.extra_info.mip_abs_gap)
        self.assertIsNotNone(results.extra_info.mip_rel_gap)
        self.assertIsNotNone(results.extra_info.mip_number_solves)

        # Check that MIP-specific fields are of correct type
        self.assertIsInstance(results.extra_info.mip_number_nodes, int)
        self.assertIsInstance(results.extra_info.mip_abs_gap, float)
        self.assertIsInstance(results.extra_info.mip_rel_gap, float)
        self.assertIsInstance(results.extra_info.mip_number_solves, int)

        # Check that non-MIP field does not exist
        self.assertFalse(hasattr(results.extra_info, 'number_iters'))

    def test_results_extra_info_no_mip(self):
        """Test that iteration info is populated for non-MIP problems."""
        opt = KnitroDirectSolver()
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(
            expr=(1.0 - m.x) ** 2 + 100.0 * (m.y - m.x**2) ** 2, sense=pyo.minimize
        )

        results = opt.solve(m)

        # Check that num_iters is populated for non-MIP
        self.assertIsNotNone(results.extra_info.number_iters)
        self.assertIsInstance(results.extra_info.number_iters, int)
        self.assertGreater(results.extra_info.number_iters, 0)

        # Check that MIP-specific fields do not exist
        self.assertFalse(hasattr(results.extra_info, 'mip_number_nodes'))
        self.assertFalse(hasattr(results.extra_info, 'mip_abs_gap'))
        self.assertFalse(hasattr(results.extra_info, 'mip_rel_gap'))
        self.assertFalse(hasattr(results.extra_info, 'mip_number_solves'))


@unittest.skipIf(not avail, "KNITRO solver is not available")
@unittest.pytest.mark.solver("knitro_direct")
class TestKnitroSolverObjectiveBound(unittest.TestCase):
    def test_objective_bound_mip(self):
        """Test that objective bound is retrieved for MIP problems."""
        opt = KnitroDirectSolver()
        m = pyo.ConcreteModel()
        m.x = pyo.Var(domain=pyo.Integers, bounds=(0, 10))
        m.y = pyo.Var(domain=pyo.Integers, bounds=(0, 10))
        m.obj = pyo.Objective(expr=m.x + m.y, sense=pyo.maximize)
        m.c1 = pyo.Constraint(expr=2 * m.x + m.y <= 15)
        m.c2 = pyo.Constraint(expr=m.x + 2 * m.y <= 15)

        results = opt.solve(m)

        # Check that objective_bound is populated
        self.assertIsNotNone(results.objective_bound)
        self.assertIsInstance(results.objective_bound, float)

        # For maximization, bound should be >= incumbent objective
        self.assertGreaterEqual(results.objective_bound, results.incumbent_objective)

    def test_objective_bound_no_mip(self):
        """Test that objective bound is not set for non-MIP problems."""
        opt = KnitroDirectSolver()
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(
            expr=(1.0 - m.x) ** 2 + 100.0 * (m.y - m.x**2) ** 2, sense=pyo.minimize
        )

        results = opt.solve(m)

        # Check that objective_bound is None for non-MIP
        self.assertIsNone(results.objective_bound)


@unittest.skipIf(not avail, "KNITRO solver is not available")
@unittest.pytest.mark.solver("knitro_direct")
class TestKnitroSolverIncumbentObjective(unittest.TestCase):
    def test_none_without_objective(self):
        """Test that incumbent objective is None when no objective is present."""
        opt = KnitroDirectSolver()
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.c1 = pyo.Constraint(expr=m.x + m.y >= 1)

        results = opt.solve(m)
        self.assertIsNone(results.incumbent_objective)

    def test_none_when_infeasible(self):
        """Test that incumbent objective is None when problem is infeasible."""
        opt = KnitroDirectSolver()
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(expr=m.x + m.y)
        m.c1 = pyo.Constraint(expr=m.x + m.y <= -20)

        opt.config.raise_exception_on_nonoptimal_result = False
        opt.config.load_solutions = False
        results = opt.solve(m)
        print(results.solution_status)
        print(results.termination_condition)
        self.assertIsNone(results.incumbent_objective)

    def test_none_when_unbounded(self):
        """Test that incumbent objective is None when problem is unbounded."""
        opt = KnitroDirectSolver()
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, None))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, None))
        m.obj = pyo.Objective(expr=m.x + m.y, sense=pyo.maximize)
        m.c1 = pyo.Constraint(expr=m.x + m.y >= -10)

        opt.config.raise_exception_on_nonoptimal_result = False
        opt.config.load_solutions = False
        results = opt.solve(m)
        self.assertIsNone(results.incumbent_objective)

    def test_value_when_optimal(self):
        """Test that incumbent objective is correct for optimal solution."""
        opt = KnitroDirectSolver()
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(expr=m.x + m.y, sense=pyo.maximize)
        m.c1 = pyo.Constraint(expr=m.x + m.y <= 3)

        results = opt.solve(m)
        self.assertIsNotNone(results.incumbent_objective)
        self.assertAlmostEqual(results.incumbent_objective, 3.0)


@unittest.skipIf(not avail, "KNITRO solver is not available")
@unittest.pytest.mark.solver("knitro_direct")
class TestKnitroSolverSolutionStatus(unittest.TestCase):
    def test_solution_status_mapping(self):
        """Test that solution status is correctly mapped from KNITRO status."""
        engine = Engine()

        # Test that RuntimeError is raised for None status
        engine._status = None
        with self.assertRaises(RuntimeError):
            engine.get_solution_status()

        # Test optimal statuses
        for code in [
            knitro.KN_RC_OPTIMAL,
            knitro.KN_RC_OPTIMAL_OR_SATISFACTORY,
            knitro.KN_RC_NEAR_OPT,
        ]:
            engine._status = code
            self.assertEqual(engine.get_solution_status(), SolutionStatus.optimal)

        # Test feasible statuses
        for code in [
            knitro.KN_RC_FEAS_XTOL,
            knitro.KN_RC_FEAS_NO_IMPROVE,
            knitro.KN_RC_FEAS_FTOL,
            -103,  # KN_RC_FEAS_BEST
            -104,  # KN_RC_FEAS_MULTISTART
            knitro.KN_RC_ITER_LIMIT_FEAS,
            knitro.KN_RC_TIME_LIMIT_FEAS,
            knitro.KN_RC_FEVAL_LIMIT_FEAS,
            knitro.KN_RC_MIP_EXH_FEAS,
            knitro.KN_RC_MIP_TERM_FEAS,
            knitro.KN_RC_MIP_SOLVE_LIMIT_FEAS,
            knitro.KN_RC_MIP_NODE_LIMIT_FEAS,
        ]:
            engine._status = code
            self.assertEqual(engine.get_solution_status(), SolutionStatus.feasible)

        # Test infeasible statuses
        for code in [
            knitro.KN_RC_INFEASIBLE,
            knitro.KN_RC_INFEAS_CON_BOUNDS,
            knitro.KN_RC_INFEAS_VAR_BOUNDS,
            knitro.KN_RC_INFEAS_MULTISTART,
        ]:
            engine._status = code
            self.assertEqual(engine.get_solution_status(), SolutionStatus.infeasible)

        # Test noSolution statuses (anything else)
        for code in [-501, -600, -99999, -1]:
            engine._status = code
            self.assertEqual(engine.get_solution_status(), SolutionStatus.noSolution)


@unittest.skipIf(not avail, "KNITRO solver is not available")
@unittest.pytest.mark.solver("knitro_direct")
class TestKnitroSolverTerminationCondition(unittest.TestCase):
    def test_termination_condition_mapping(self):
        """Test that termination condition is correctly mapped from KNITRO status."""
        engine = Engine()

        # Test that RuntimeError is raised for None status
        engine._status = None
        with self.assertRaises(RuntimeError):
            engine.get_termination_condition()

        # Test convergenceCriteriaSatisfied
        for code in [
            knitro.KN_RC_OPTIMAL,
            knitro.KN_RC_OPTIMAL_OR_SATISFACTORY,
            knitro.KN_RC_NEAR_OPT,
        ]:
            engine._status = code
            self.assertEqual(
                engine.get_termination_condition(),
                TerminationCondition.convergenceCriteriaSatisfied,
            )

        # Test locallyInfeasible
        for code in [knitro.KN_RC_INFEAS_NO_IMPROVE, knitro.KN_RC_INFEAS_MULTISTART]:
            engine._status = code
            self.assertEqual(
                engine.get_termination_condition(),
                TerminationCondition.locallyInfeasible,
            )

        # Test provenInfeasible
        for code in [
            knitro.KN_RC_INFEASIBLE,
            knitro.KN_RC_INFEAS_CON_BOUNDS,
            knitro.KN_RC_INFEAS_VAR_BOUNDS,
        ]:
            engine._status = code
            self.assertEqual(
                engine.get_termination_condition(),
                TerminationCondition.provenInfeasible,
            )

        # Test infeasibleOrUnbounded
        for code in [knitro.KN_RC_UNBOUNDED, knitro.KN_RC_UNBOUNDED_OR_INFEAS]:
            engine._status = code
            self.assertEqual(
                engine.get_termination_condition(),
                TerminationCondition.infeasibleOrUnbounded,
            )

        # Test iterationLimit
        for code in [
            knitro.KN_RC_ITER_LIMIT_FEAS,
            knitro.KN_RC_FEVAL_LIMIT_FEAS,
            knitro.KN_RC_MIP_EXH_FEAS,
            knitro.KN_RC_MIP_TERM_FEAS,
            knitro.KN_RC_MIP_SOLVE_LIMIT_FEAS,
            knitro.KN_RC_MIP_NODE_LIMIT_FEAS,
            knitro.KN_RC_ITER_LIMIT_INFEAS,
            knitro.KN_RC_FEVAL_LIMIT_INFEAS,
            knitro.KN_RC_MIP_EXH_INFEAS,
            knitro.KN_RC_MIP_SOLVE_LIMIT_INFEAS,
            knitro.KN_RC_MIP_NODE_LIMIT_INFEAS,
        ]:
            engine._status = code
            self.assertEqual(
                engine.get_termination_condition(), TerminationCondition.iterationLimit
            )

        # Test interrupted
        engine._status = knitro.KN_RC_USER_TERMINATION
        self.assertEqual(
            engine.get_termination_condition(), TerminationCondition.interrupted
        )

        # Test error (-500 > status >= -600)
        for code in [-501, -550, -600]:
            engine._status = code
            self.assertEqual(
                engine.get_termination_condition(), TerminationCondition.error
            )

        # Test unknown (anything else outside the defined ranges)
        for code in [-601, -99999, -1]:
            engine._status = code
            self.assertEqual(
                engine.get_termination_condition(), TerminationCondition.unknown
            )


@unittest.skipIf(not avail, "KNITRO solver is not available")
@unittest.pytest.mark.solver("knitro_direct")
class TestKnitroDirectSolverInterface(unittest.TestCase):
    def test_class_member_list(self):
        opt = KnitroDirectSolver()
        expected_list = [
            "CONFIG",
            "available",
            "config",
            "api_version",
            "is_persistent",
            "name",
            "solve",
            "version",
        ]
        method_list = [
            m for m in dir(opt) if not m.startswith("_") and not m.startswith("get")
        ]
        self.assertListEqual(sorted(method_list), sorted(expected_list))

    def test_default_instantiation(self):
        opt = KnitroDirectSolver()
        self.assertFalse(opt.is_persistent())
        self.assertIsNotNone(opt.version())
        self.assertEqual(opt.name, "knitro_direct")
        self.assertEqual(opt.CONFIG, opt.config)
        self.assertTrue(opt.available())

    def test_instantiation_as_context(self):
        with KnitroDirectSolver() as opt:
            self.assertFalse(opt.is_persistent())
            self.assertIsNotNone(opt.version())
            self.assertEqual(opt.name, "knitro_direct")
            self.assertEqual(opt.CONFIG, opt.config)
            self.assertTrue(opt.available())

    def test_available_cache(self):
        opt = KnitroDirectSolver()
        opt.available()
        self.assertTrue(opt._available_cache)
        self.assertIsNotNone(opt._available_cache)


@unittest.skipIf(not avail, "KNITRO solver is not available")
@unittest.pytest.mark.solver("knitro_direct")
class TestKnitroDirectSolver(unittest.TestCase):
    def setUp(self):
        self.opt = KnitroDirectSolver()

    def test_solve_tee(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(
            expr=(1.0 - m.x) + 100.0 * (m.y - m.x), sense=pyo.minimize
        )
        stream = io.StringIO()
        with redirect_stdout(stream):
            self.opt.solve(m, tee=True)
        output = stream.getvalue()
        self.assertTrue(bool(output.strip()))

    def test_solve_no_tee(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(
            expr=(1.0 - m.x) + 100.0 * (m.y - m.x), sense=pyo.minimize
        )
        stream = io.StringIO()
        with redirect_stdout(stream):
            self.opt.solve(m, tee=False)
        output = stream.getvalue()
        self.assertFalse(bool(output.strip()))

    def test_solve(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(
            expr=(1.0 - m.x) + 100.0 * (m.y - m.x), sense=pyo.minimize
        )
        res = self.opt.solve(m)
        self.assertAlmostEqual(res.incumbent_objective, -1004)
        self.assertAlmostEqual(m.x.value, 5)
        self.assertAlmostEqual(m.y.value, -5)

    def test_qp_solve(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(
            expr=(1.0 - m.x) + 100.0 * (m.y - m.x) ** 2, sense=pyo.minimize
        )
        results = self.opt.solve(m)
        self.assertAlmostEqual(results.incumbent_objective, -4.0, 3)
        self.assertAlmostEqual(m.x.value, 5.0, 3)
        self.assertAlmostEqual(m.y.value, 5.0, 3)

    def test_qcp_solve(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.obj = pyo.Objective(expr=(m.y - m.x) ** 2, sense=pyo.minimize)
        m.c1 = pyo.Constraint(expr=m.x**2 + m.y**2 <= 4)
        results = self.opt.solve(m)
        self.assertAlmostEqual(results.incumbent_objective, 0.0)

    def test_solve_exp(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.y >= pyo.exp(m.x))
        self.opt.solve(m)
        self.assertAlmostEqual(m.x.value, -0.42630274815985264)
        self.assertAlmostEqual(m.y.value, 0.6529186341994245)

    def test_solve_log(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1)
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.y <= pyo.log(m.x))
        self.opt.solve(m)
        self.assertAlmostEqual(m.x.value, 0.6529186341994245)
        self.assertAlmostEqual(m.y.value, -0.42630274815985264)

    def test_solve_HS071(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(pyo.RangeSet(1, 4), bounds=(1.0, 5.0))
        m.obj = pyo.Objective(
            expr=m.x[1] * m.x[4] * (m.x[1] + m.x[2] + m.x[3]) + m.x[3],
            sense=pyo.minimize,
        )
        m.c1 = pyo.Constraint(expr=m.x[1] * m.x[2] * m.x[3] * m.x[4] >= 25.0)
        m.c2 = pyo.Constraint(
            expr=m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2 + m.x[4] ** 2 == 40.0
        )
        self.opt.solve(m, solver_options={"opttol": 1e-5})
        self.assertAlmostEqual(pyo.value(m.x[1]), 1.0, 3)
        self.assertAlmostEqual(pyo.value(m.x[2]), 4.743, 3)
        self.assertAlmostEqual(pyo.value(m.x[3]), 3.821, 3)
        self.assertAlmostEqual(pyo.value(m.x[4]), 1.379, 3)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.piecewise.tests import models
import pyomo.contrib.piecewise.tests.common_tests as ct
from pyomo.core.base import TransformationFactory
from pyomo.environ import SolverFactory, Var, Constraint
from pyomo.core.expr.compare import assertExpressionsEqual


class TestTransformPiecewiseModelToNestedInnerRepnMIP(unittest.TestCase):
    def check_pw_log(self, m):
        z = m.pw_log.get_transformation_var(m.log_expr)
        self.assertIsInstance(z, Var)
        # Now we can use those Vars to check on what the transformation created
        log_block = z.parent_block()

        # We should have three Vars, two of which are indexed, and five
        # Constraints, three of which are indexed

        self.assertEqual(len(log_block.component_map(Var)), 3)
        self.assertEqual(len(log_block.component_map(Constraint)), 5)

        # Constants
        simplex_count = 3
        log_simplex_count = 2
        simplex_point_count = 2

        # Substitute var
        self.assertIsInstance(log_block.substitute_var, Var)
        self.assertIs(m.obj.expr.expr, log_block.substitute_var)
        # Binaries
        self.assertIsInstance(log_block.binaries, Var)
        self.assertEqual(len(log_block.binaries), log_simplex_count)
        # Lambdas
        self.assertIsInstance(log_block.lambdas, Var)
        self.assertEqual(len(log_block.lambdas), simplex_count * simplex_point_count)
        for l in log_block.lambdas.values():
            self.assertEqual(l.lb, 0)
            self.assertEqual(l.ub, 1)

        # Convex combo constraint
        self.assertIsInstance(log_block.convex_combo, Constraint)
        assertExpressionsEqual(
            self,
            log_block.convex_combo.expr,
            log_block.lambdas[0, 0]
            + log_block.lambdas[0, 1]
            + log_block.lambdas[1, 0]
            + log_block.lambdas[1, 1]
            + log_block.lambdas[2, 0]
            + log_block.lambdas[2, 1]
            == 1,
        )

        # Set substitute constraint
        self.assertIsInstance(log_block.set_substitute, Constraint)
        assertExpressionsEqual(
            self,
            log_block.set_substitute.expr,
            log_block.substitute_var
            == log_block.lambdas[0, 0] * m.f1(1)
            + log_block.lambdas[1, 0] * m.f2(3)
            + log_block.lambdas[2, 0] * m.f3(6)
            + log_block.lambdas[0, 1] * m.f1(3)
            + log_block.lambdas[1, 1] * m.f2(6)
            + log_block.lambdas[2, 1] * m.f3(10),
            places=7,
        )

        # x constraint
        self.assertIsInstance(log_block.x_constraint, Constraint)
        # one-dimensional case, so there is only one x variable here
        self.assertEqual(len(log_block.x_constraint), 1)
        assertExpressionsEqual(
            self,
            log_block.x_constraint[0].expr,
            m.x
            == 1 * log_block.lambdas[0, 0]
            + 3 * log_block.lambdas[0, 1]
            + 3 * log_block.lambdas[1, 0]
            + 6 * log_block.lambdas[1, 1]
            + 6 * log_block.lambdas[2, 0]
            + 10 * log_block.lambdas[2, 1],
        )

        # simplex choice 1 constraint enables lambdas when binaries are on
        self.assertEqual(len(log_block.simplex_choice_1), log_simplex_count)
        assertExpressionsEqual(
            self,
            log_block.simplex_choice_1[0].expr,
            log_block.lambdas[2, 0] + log_block.lambdas[2, 1] <= log_block.binaries[0],
        )
        assertExpressionsEqual(
            self,
            log_block.simplex_choice_1[1].expr,
            log_block.lambdas[1, 0] + log_block.lambdas[1, 1] <= log_block.binaries[1],
        )
        # simplex choice 2 constraint enables lambdas when binaries are off
        self.assertEqual(len(log_block.simplex_choice_2), log_simplex_count)
        assertExpressionsEqual(
            self,
            log_block.simplex_choice_2[0].expr,
            log_block.lambdas[0, 0]
            + log_block.lambdas[0, 1]
            + log_block.lambdas[1, 0]
            + log_block.lambdas[1, 1]
            <= 1 - log_block.binaries[0],
        )
        assertExpressionsEqual(
            self,
            log_block.simplex_choice_2[1].expr,
            log_block.lambdas[0, 0]
            + log_block.lambdas[0, 1]
            + log_block.lambdas[2, 0]
            + log_block.lambdas[2, 1]
            <= 1 - log_block.binaries[1],
        )

    def check_pw_paraboloid(self, m):
        # This is a little larger, but at least test that the right numbers of
        # everything are created
        z = m.pw_paraboloid.get_transformation_var(m.paraboloid_expr)
        self.assertIsInstance(z, Var)
        paraboloid_block = z.parent_block()

        self.assertEqual(len(paraboloid_block.component_map(Var)), 3)
        self.assertEqual(len(paraboloid_block.component_map(Constraint)), 5)

        # Constants
        simplex_count = 4
        log_simplex_count = 2
        simplex_point_count = 3

        # Substitute var
        self.assertIsInstance(paraboloid_block.substitute_var, Var)
        # Binaries
        self.assertIsInstance(paraboloid_block.binaries, Var)
        self.assertEqual(len(paraboloid_block.binaries), log_simplex_count)
        # Lambdas
        self.assertIsInstance(paraboloid_block.lambdas, Var)
        self.assertEqual(
            len(paraboloid_block.lambdas), simplex_count * simplex_point_count
        )
        for l in paraboloid_block.lambdas.values():
            self.assertEqual(l.lb, 0)
            self.assertEqual(l.ub, 1)

        # Convex combo constraint
        self.assertIsInstance(paraboloid_block.convex_combo, Constraint)
        assertExpressionsEqual(
            self,
            paraboloid_block.convex_combo.expr,
            paraboloid_block.lambdas[0, 0]
            + paraboloid_block.lambdas[0, 1]
            + paraboloid_block.lambdas[0, 2]
            + paraboloid_block.lambdas[1, 0]
            + paraboloid_block.lambdas[1, 1]
            + paraboloid_block.lambdas[1, 2]
            + paraboloid_block.lambdas[2, 0]
            + paraboloid_block.lambdas[2, 1]
            + paraboloid_block.lambdas[2, 2]
            + paraboloid_block.lambdas[3, 0]
            + paraboloid_block.lambdas[3, 1]
            + paraboloid_block.lambdas[3, 2]
            == 1,
        )

        # Set substitute constraint
        self.assertIsInstance(paraboloid_block.set_substitute, Constraint)
        assertExpressionsEqual(
            self,
            paraboloid_block.set_substitute.expr,
            paraboloid_block.substitute_var
            == paraboloid_block.lambdas[0, 0] * m.g1(0, 1)
            + paraboloid_block.lambdas[1, 0] * m.g1(0, 1)
            + paraboloid_block.lambdas[2, 0] * m.g2(3, 4)
            + paraboloid_block.lambdas[3, 0] * m.g2(0, 7)
            + paraboloid_block.lambdas[0, 1] * m.g1(0, 4)
            + paraboloid_block.lambdas[1, 1] * m.g1(3, 4)
            + paraboloid_block.lambdas[2, 1] * m.g2(3, 7)
            + paraboloid_block.lambdas[3, 1] * m.g2(0, 4)
            + paraboloid_block.lambdas[0, 2] * m.g1(3, 4)
            + paraboloid_block.lambdas[1, 2] * m.g1(3, 1)
            + paraboloid_block.lambdas[2, 2] * m.g2(0, 7)
            + paraboloid_block.lambdas[3, 2] * m.g2(3, 4),
            places=7,
        )

        # x constraint
        self.assertIsInstance(paraboloid_block.x_constraint, Constraint)
        # Here we have two x variables
        self.assertEqual(len(paraboloid_block.x_constraint), 2)
        assertExpressionsEqual(
            self,
            paraboloid_block.x_constraint[0].expr,
            m.x1
            == 0 * paraboloid_block.lambdas[0, 0]
            + 0 * paraboloid_block.lambdas[0, 1]
            + 3 * paraboloid_block.lambdas[0, 2]
            + 0 * paraboloid_block.lambdas[1, 0]
            + 3 * paraboloid_block.lambdas[1, 1]
            + 3 * paraboloid_block.lambdas[1, 2]
            + 3 * paraboloid_block.lambdas[2, 0]
            + 3 * paraboloid_block.lambdas[2, 1]
            + 0 * paraboloid_block.lambdas[2, 2]
            + 0 * paraboloid_block.lambdas[3, 0]
            + 0 * paraboloid_block.lambdas[3, 1]
            + 3 * paraboloid_block.lambdas[3, 2],
        )
        assertExpressionsEqual(
            self,
            paraboloid_block.x_constraint[1].expr,
            m.x2
            == 1 * paraboloid_block.lambdas[0, 0]
            + 4 * paraboloid_block.lambdas[0, 1]
            + 4 * paraboloid_block.lambdas[0, 2]
            + 1 * paraboloid_block.lambdas[1, 0]
            + 4 * paraboloid_block.lambdas[1, 1]
            + 1 * paraboloid_block.lambdas[1, 2]
            + 4 * paraboloid_block.lambdas[2, 0]
            + 7 * paraboloid_block.lambdas[2, 1]
            + 7 * paraboloid_block.lambdas[2, 2]
            + 7 * paraboloid_block.lambdas[3, 0]
            + 4 * paraboloid_block.lambdas[3, 1]
            + 4 * paraboloid_block.lambdas[3, 2],
        )

        # The choices will get long, so let's just assert we have enough
        self.assertEqual(len(paraboloid_block.simplex_choice_1), log_simplex_count)
        self.assertEqual(len(paraboloid_block.simplex_choice_2), log_simplex_count)

    # Test methods using the common_tests.py code.
    def test_transformation_do_not_descend(self):
        ct.check_transformation_do_not_descend(
            self, 'contrib.piecewise.disaggregated_logarithmic'
        )

    def test_transformation_PiecewiseLinearFunction_targets(self):
        ct.check_transformation_PiecewiseLinearFunction_targets(
            self, 'contrib.piecewise.disaggregated_logarithmic'
        )

    def test_descend_into_expressions(self):
        ct.check_descend_into_expressions(
            self, 'contrib.piecewise.disaggregated_logarithmic'
        )

    def test_descend_into_expressions_constraint_target(self):
        ct.check_descend_into_expressions_constraint_target(
            self, 'contrib.piecewise.disaggregated_logarithmic'
        )

    def test_descend_into_expressions_objective_target(self):
        ct.check_descend_into_expressions_objective_target(
            self, 'contrib.piecewise.disaggregated_logarithmic'
        )

    # Check solution of the log(x) model
    @unittest.skipUnless(SolverFactory('gurobi').available(), 'Gurobi is not available')
    @unittest.skipUnless(SolverFactory('gurobi').license_is_valid(), 'No license')
    def test_solve_log_model(self):
        m = models.make_log_x_model()
        TransformationFactory("contrib.piecewise.disaggregated_logarithmic").apply_to(m)
        SolverFactory("gurobi").solve(m)
        ct.check_log_x_model_soln(self, m)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.common.tempfiles import TempfileManager

from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)
from pyomo.common.dependencies.scipy import sparse as spa

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")

from pyomo.contrib.pynumero.exceptions import PyNumeroEvaluationError
from pyomo.contrib.pynumero.asl import AmplInterface

if not AmplInterface.available():
    raise unittest.SkipTest(
        "Pynumero needs the ASL extension to run CyIpoptSolver tests"
    )

from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP

from pyomo.contrib.pynumero.interfaces.cyipopt_interface import (
    cyipopt,
    cyipopt_available,
    CyIpoptNLP,
)

from pyomo.contrib.pynumero.algorithms.solvers.cyipopt_solver import CyIpoptSolver

if cyipopt_available:
    # We don't raise unittest.SkipTest if not cyipopt_available as there is a
    # test below that tests an exception when cyipopt is unavailable.
    cyipopt_ge_1_3 = hasattr(cyipopt, "CyIpoptEvaluationError")
    ipopt_ge_3_14 = cyipopt.IPOPT_VERSION >= (3, 14, 0)


def create_model1():
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2, 3], initialize=4.0)
    m.c = pyo.Constraint(expr=m.x[3] ** 2 + m.x[1] == 25)
    m.d = pyo.Constraint(expr=m.x[2] ** 2 + m.x[1] <= 18.0)
    m.o = pyo.Objective(expr=m.x[1] ** 4 - 3 * m.x[1] * m.x[2] ** 3 + m.x[3] ** 2 - 8.0)
    m.x[1].setlb(0.0)
    m.x[2].setlb(0.0)

    return m


def create_model2():
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2], initialize=4.0)
    m.d = pyo.Constraint(expr=m.x[1] + m.x[2] <= 5)
    m.o = pyo.Objective(expr=m.x[1] ** 2 + 4 * m.x[2] ** 2 - 8 * m.x[1] - 16 * m.x[2])
    m.x[1].setub(3.0)
    m.x[1].setlb(0.0)
    m.x[2].setlb(0.0)

    return m


def create_model3(G, A, b, c):
    nx = G.shape[0]
    nl = A.shape[0]

    model = pyo.ConcreteModel()
    model.var_ids = range(nx)
    model.con_ids = range(nl)

    model.x = pyo.Var(model.var_ids, initialize=0.0)
    model.hessian_f = pyo.Param(
        model.var_ids, model.var_ids, mutable=True, rule=lambda m, i, j: G[i, j]
    )
    model.jacobian_c = pyo.Param(
        model.con_ids, model.var_ids, mutable=True, rule=lambda m, i, j: A[i, j]
    )
    model.rhs = pyo.Param(model.con_ids, mutable=True, rule=lambda m, i: b[i])
    model.grad_f = pyo.Param(model.var_ids, mutable=True, rule=lambda m, i: c[i])

    def equality_constraint_rule(m, i):
        return sum(m.jacobian_c[i, j] * m.x[j] for j in m.var_ids) == m.rhs[i]

    model.equalities = pyo.Constraint(model.con_ids, rule=equality_constraint_rule)

    def objective_rule(m):
        accum = 0.0
        for i in m.var_ids:
            accum += m.x[i] * sum(m.hessian_f[i, j] * m.x[j] for j in m.var_ids)
        accum *= 0.5
        accum += sum(m.x[j] * m.grad_f[j] for j in m.var_ids)
        return accum

    model.obj = pyo.Objective(rule=objective_rule, sense=pyo.minimize)

    return model


def create_model4():
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2], initialize=1.0)
    m.c1 = pyo.Constraint(expr=m.x[1] + m.x[2] - 1 == 0)
    m.obj = pyo.Objective(expr=2 * m.x[1] ** 2 + m.x[2] ** 2)
    return m


def create_model6():
    model = pyo.ConcreteModel()

    model.S = [1, 2]
    model.x = pyo.Var(model.S, initialize=1.0)

    def f(model):
        return (
            model.x[1] ** 4
            + (model.x[1] + model.x[2]) ** 2
            + (-1.0 + pyo.exp(model.x[2])) ** 2
        )

    model.f = pyo.Objective(rule=f)
    return model


def create_model9():
    # clplatea OXR2-MN-V-0
    model = pyo.ConcreteModel()

    p = 71
    wght = -0.1
    hp2 = 0.5 * p**2

    model.x = pyo.Var(pyo.RangeSet(1, p), pyo.RangeSet(1, p), initialize=0.0)

    def f(model):
        return sum(
            0.5 * (model.x[i, j] - model.x[i, j - 1]) ** 2
            + 0.5 * (model.x[i, j] - model.x[i - 1, j]) ** 2
            + hp2 * (model.x[i, j] - model.x[i, j - 1]) ** 4
            + hp2 * (model.x[i, j] - model.x[i - 1, j]) ** 4
            for i in range(2, p + 1)
            for j in range(2, p + 1)
        ) + (wght * model.x[p, p])

    model.f = pyo.Objective(rule=f)

    for j in range(1, p + 1):
        model.x[1, j] = 0.0
        model.x[1, j].fixed = True

    return model


def make_hs071_model():
    # This is a model that is mathematically equivalent to the Hock-Schittkowski
    # test problem 071, but that will trigger an evaluation error if x[0] goes
    # above 1.1.
    m = pyo.ConcreteModel()
    m.x = pyo.Var([0, 1, 2, 3], bounds=(1.0, 5.0))
    m.x[0] = 1.0
    m.x[1] = 5.0
    m.x[2] = 5.0
    m.x[3] = 1.0
    m.obj = pyo.Objective(expr=m.x[0] * m.x[3] * (m.x[0] + m.x[1] + m.x[2]) + m.x[2])
    # This expression evaluates to zero, but is not well defined when x[0] > 1.1
    trivial_expr_with_eval_error = (pyo.sqrt(1.1 - m.x[0])) ** 2 + m.x[0] - 1.1
    m.ineq1 = pyo.Constraint(expr=m.x[0] * m.x[1] * m.x[2] * m.x[3] >= 25.0)
    m.eq1 = pyo.Constraint(
        expr=(
            m.x[0] ** 2 + m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2
            == 40.0 + trivial_expr_with_eval_error
        )
    )
    return m


@unittest.skipIf(cyipopt_available, "cyipopt is available")
class TestCyIpoptNotAvailable(unittest.TestCase):
    def test_not_available_exception(self):
        model = create_model1()
        nlp = PyomoNLP(model)
        msg = "cyipopt is required"
        with self.assertRaisesRegex(RuntimeError, msg):
            solver = CyIpoptSolver(CyIpoptNLP(nlp))


@unittest.skipUnless(cyipopt_available, "cyipopt is not available")
class TestCyIpoptSolver(unittest.TestCase):
    def test_model1(self):
        model = create_model1()
        nlp = PyomoNLP(model)
        solver = CyIpoptSolver(CyIpoptNLP(nlp))
        x, info = solver.solve(tee=False)
        x_sol = np.array([3.85958688, 4.67936007, 3.10358931])
        y_sol = np.array([-1.0, 53.90357665])
        self.assertTrue(np.allclose(x, x_sol, rtol=1e-4))
        nlp.set_primals(x)
        nlp.set_duals(y_sol)
        self.assertAlmostEqual(nlp.evaluate_objective(), -428.6362455416348, places=5)
        self.assertTrue(np.allclose(info['mult_g'], y_sol, rtol=1e-4))

    def test_model1_with_scaling(self):
        m = create_model1()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        m.scaling_factor[m.o] = 1e-6  # scale the objective
        m.scaling_factor[m.c] = 2.0  # scale the equality constraint
        m.scaling_factor[m.d] = 3.0  # scale the inequality constraint
        m.scaling_factor[m.x[1]] = 4.0  # scale one of the x variables

        with TempfileManager.new_context() as temp:
            cynlp = CyIpoptNLP(PyomoNLP(m))
            logfile = temp.create_tempfile('_cyipopt-scaling.log')
            options = {
                'nlp_scaling_method': 'user-scaling',
                'output_file': logfile,
                'file_print_level': 10,
                'max_iter': 0,
            }
            solver = CyIpoptSolver(cynlp, options=options)
            x, info = solver.solve()
            cynlp.close()

            with open(logfile, 'r') as fd:
                solver_trace = fd.read()

        # check for the following strings in the log
        self.assertIn('nlp_scaling_method = user-scaling', solver_trace)
        self.assertIn(f"output_file = {logfile}", solver_trace)
        self.assertIn('objective scaling factor = 1e-06', solver_trace)
        self.assertIn('x scaling provided', solver_trace)
        self.assertIn('c scaling provided', solver_trace)
        self.assertIn('d scaling provided', solver_trace)
        self.assertIn('DenseVector "x scaling vector" with 3 elements:', solver_trace)
        self.assertIn('x scaling vector[    1]= 1.0000000000000000e+00', solver_trace)
        self.assertIn('x scaling vector[    2]= 1.0000000000000000e+00', solver_trace)
        self.assertIn('x scaling vector[    3]= 4.0000000000000000e+00', solver_trace)
        self.assertIn('DenseVector "c scaling vector" with 1 elements:', solver_trace)
        self.assertIn('c scaling vector[    1]= 2.0000000000000000e+00', solver_trace)
        self.assertIn('DenseVector "d scaling vector" with 1 elements:', solver_trace)
        self.assertIn('d scaling vector[    1]= 3.0000000000000000e+00', solver_trace)

    def test_model2(self):
        model = create_model2()
        nlp = PyomoNLP(model)
        solver = CyIpoptSolver(CyIpoptNLP(nlp))
        x, info = solver.solve(tee=False)
        x_sol = np.array([3.0, 1.99997807])
        y_sol = np.array([0.00017543])
        self.assertTrue(np.allclose(x, x_sol, rtol=1e-4))
        nlp.set_primals(x)
        nlp.set_duals(y_sol)
        self.assertAlmostEqual(nlp.evaluate_objective(), -31.000000057167462, places=5)
        self.assertTrue(np.allclose(info['mult_g'], y_sol, rtol=1e-4))

    def test_model3(self):
        G = np.array([[6, 2, 1], [2, 5, 2], [1, 2, 4]])
        A = np.array([[1, 0, 1], [0, 1, 1]])
        b = np.array([3, 0])
        c = np.array([-8, -3, -3])

        model = create_model3(G, A, b, c)
        nlp = PyomoNLP(model)
        solver = CyIpoptSolver(CyIpoptNLP(nlp))
        x, info = solver.solve(tee=False)
        x_sol = np.array([2.0, -1.0, 1.0])
        y_sol = np.array([-3.0, 2.0])
        self.assertTrue(np.allclose(x, x_sol, rtol=1e-4))
        nlp.set_primals(x)
        nlp.set_duals(y_sol)
        self.assertAlmostEqual(nlp.evaluate_objective(), -3.5, places=5)
        self.assertTrue(np.allclose(info['mult_g'], y_sol, rtol=1e-4))

    def test_options(self):
        model = create_model1()
        nlp = PyomoNLP(model)
        solver = CyIpoptSolver(CyIpoptNLP(nlp), options={'max_iter': 1})
        x, info = solver.solve(tee=False)
        nlp.set_primals(x)
        self.assertAlmostEqual(nlp.evaluate_objective(), -5.0879028e02, places=5)

    @unittest.skipUnless(
        cyipopt_available and cyipopt_ge_1_3, "cyipopt version < 1.3.0"
    )
    def test_hs071_evalerror(self):
        m = make_hs071_model()
        solver = pyo.SolverFactory("cyipopt")
        res = solver.solve(m, tee=True)

        x = list(m.x[:].value)
        expected_x = np.array([1.0, 4.74299964, 3.82114998, 1.37940829])
        np.testing.assert_allclose(x, expected_x)

    def test_hs071_evalerror_halt(self):
        m = make_hs071_model()
        solver = pyo.SolverFactory("cyipopt", halt_on_evaluation_error=True)
        msg = "Error in AMPL evaluation"
        with self.assertRaisesRegex(PyNumeroEvaluationError, msg):
            res = solver.solve(m, tee=True)

    @unittest.skipIf(
        not cyipopt_available or cyipopt_ge_1_3, "cyipopt version >= 1.3.0"
    )
    def test_hs071_evalerror_old_cyipopt(self):
        m = make_hs071_model()
        solver = pyo.SolverFactory("cyipopt")
        msg = "Error in AMPL evaluation"
        with self.assertRaisesRegex(PyNumeroEvaluationError, msg):
            res = solver.solve(m, tee=True)

    def test_solve_without_objective(self):
        m = create_model1()
        m.o.deactivate()
        m.x[2].fix(0.0)
        m.x[3].fix(4.0)
        solver = pyo.SolverFactory("cyipopt")
        res = solver.solve(m, tee=True)
        pyo.assert_optimal_termination(res)
        self.assertAlmostEqual(m.x[1].value, 9.0)

    def test_solve_13arg_callback(self):
        m = create_model1()

        iterate_data = []

        def intermediate(
            nlp,
            alg_mod,
            iter_count,
            obj_value,
            inf_pr,
            inf_du,
            mu,
            d_norm,
            regularization_size,
            alpha_du,
            alpha_pr,
            ls_trials,
        ):
            x = nlp.get_primals()
            y = nlp.get_duals()
            iterate_data.append((x, y))

        x_sol = np.array([3.85958688, 4.67936007, 3.10358931])
        y_sol = np.array([-1.0, 53.90357665])

        solver = pyo.SolverFactory("cyipopt", intermediate_callback=intermediate)
        res = solver.solve(m, tee=True)
        pyo.assert_optimal_termination(res)

        # Make sure iterate vectors have the right shape and that the final
        # iterate contains the primal solution we expect.
        for x, y in iterate_data:
            self.assertEqual(x.shape, (3,))
            self.assertEqual(y.shape, (2,))
        x, y = iterate_data[-1]
        self.assertTrue(np.allclose(x_sol, x))
        # Note that we can't assert that dual variables in the NLP are those
        # at the solution because, at this point in the algorithm, the NLP
        # only has access to the *previous iteration's* dual values.

    # The 13-arg callback works with cyipopt < 1.3, but we will use the
    # get_current_iterate method, which is only available in 1.3+ and IPOPT 3.14+
    @unittest.skipIf(
        not cyipopt_available or not cyipopt_ge_1_3 or not ipopt_ge_3_14,
        "cyipopt version < 1.3.0",
    )
    def test_solve_get_current_iterate(self):
        m = create_model1()

        iterate_data = []

        def intermediate(
            nlp,
            problem,
            alg_mod,
            iter_count,
            obj_value,
            inf_pr,
            inf_du,
            mu,
            d_norm,
            regularization_size,
            alpha_du,
            alpha_pr,
            ls_trials,
        ):
            iterate = problem.get_current_iterate()
            x = iterate["x"]
            y = iterate["mult_g"]
            iterate_data.append((x, y))

        x_sol = np.array([3.85958688, 4.67936007, 3.10358931])
        y_sol = np.array([-1.0, 53.90357665])

        solver = pyo.SolverFactory("cyipopt", intermediate_callback=intermediate)
        res = solver.solve(m, tee=True)
        pyo.assert_optimal_termination(res)

        # Make sure iterate vectors have the right shape and that the final
        # iterate contains the primal and dual solution we expect.
        for x, y in iterate_data:
            self.assertEqual(x.shape, (3,))
            self.assertEqual(y.shape, (2,))
        x, y = iterate_data[-1]
        self.assertTrue(np.allclose(x_sol, x))
        self.assertTrue(np.allclose(y_sol, y))
